package com.untzuntz.ustackclientapi;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Type;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Scanner;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonDeserializationContext;
import com.google.gson.JsonDeserializer;
import com.google.gson.JsonElement;
import com.google.gson.JsonParseException;
import com.untzuntz.ustackclientapi.exceptions.APIConnectionException;
import com.untzuntz.ustackclientapi.exceptions.APIErrorException;
import com.untzuntz.ustackclientapi.exceptions.APIException;
import com.untzuntz.ustackclientapi.exceptions.AuthenticationException;
import com.untzuntz.ustackclientapi.exceptions.ServerSideException;
import com.untzuntz.ustackclientapi.status.Status;

public class APIRequest {

	public static final Gson gson = new GsonBuilder().registerTypeAdapter(JsonElement.class, JsonElementDeserializer.INSTANCE).setDateFormat("yyyyMMddHHmmssZ").create();
	protected static final String CHARSET = "UTF-8";
	protected enum Method { GET, POST, PUT, DELETE }

	protected static String urlEncodePair(String k, String v) throws UnsupportedEncodingException {
		return String.format("%s=%s", 
				URLEncoder.encode(k, CHARSET), 
				URLEncoder.encode(v, CHARSET));
	}
	
	public enum JsonElementDeserializer implements JsonDeserializer<JsonElement> {
		  INSTANCE;

		  public JsonElement deserialize(
		      JsonElement json, Type typeOfT, JsonDeserializationContext context) 
		      throws JsonParseException {
		    return json;
		  }
		}

	/**
	 * Returns a CSV provided an array of strings
	 * @param array
	 * @return
	 */
	public static String arrayToCsv(String[] array)
	{
		if (array == null || array.length == 0)
			return "";
		
		StringBuffer ret = new StringBuffer();

		for (int i = 0; i < array.length; i++)
		{
			String s = array[i];
			ret.append(s.trim());
			if ((i + 1) < array.length)
				ret.append(",");
		}
		
		return ret.toString();
	}
	
	private static HttpURLConnection createConnection(String url) throws IOException {
		
		URL aURL = new URL(url);
		HttpURLConnection conn = (HttpURLConnection) aURL.openConnection(); //use SSL URLs

		if (url.startsWith("https://"))
			((HttpsURLConnection)conn).setHostnameVerifier(new CustomizedHostnameVerifier());
		
		conn.setConnectTimeout(30000);
		conn.setReadTimeout(90000);
		conn.setUseCaches(false);
		return conn;

	}
	
	public static HttpURLConnection createGetConnection(String url, String query) throws IOException {
		String getURL = String.format("%s?%s", url, query);
		HttpURLConnection conn = createConnection(getURL);
		conn.setRequestMethod("GET");
		return conn;
	}

	public static HttpURLConnection createPutConnection(String url, String query) throws IOException {
		String getURL = String.format("%s?%s", url, query);
		HttpURLConnection conn = createConnection(getURL);
		conn.setRequestMethod("PUT");
		return conn;
	}

	public static HttpURLConnection createDeleteConnection(String url, String query) throws IOException {
		String getURL = String.format("%s?%s", url, query);
		HttpURLConnection conn = createConnection(getURL);
		conn.setRequestMethod("DELETE");
		return conn;
	}

	public static HttpURLConnection createPostConnection(String url, String query) throws IOException {
		URL aURL = new URL(url);
		HttpURLConnection conn = null;
		
		if (url.startsWith("https://"))
		{
			conn = (HttpsURLConnection) aURL.openConnection(); //use SSL URLs
			((HttpsURLConnection)conn).setHostnameVerifier(new CustomizedHostnameVerifier());
		}
		else
			conn = (HttpURLConnection) aURL.openConnection();
		
		conn.setConnectTimeout(30000);
		conn.setReadTimeout(90000);
		conn.setUseCaches(false);
		conn.setDoInput(true);
		conn.setDoOutput(true);
		conn.setUseCaches(false);
		conn.setRequestMethod("POST");
		
		OutputStream output = null;
		try {
			DataOutputStream httpStreamWriter = new DataOutputStream(conn.getOutputStream());
			httpStreamWriter.writeBytes(query);
			httpStreamWriter.flush ();
			httpStreamWriter.close ();
		} finally {
			if (output != null) output.close();
		}
		
		return conn;
	}
	
	public static String getResponseBody(InputStream responseStream) throws IOException {
		String rBody = new Scanner(responseStream, CHARSET).useDelimiter("\\A").next(); // \A is the beginning of the stream boundary
		responseStream.close();
		return rBody;
	}

	protected static APIResponse makeURLConnectionRequest(Method method, String url, String query) throws APIConnectionException {
		
		System.out.println("Type [" + method + "] URL [" + url + "] Query [" + query + "]");
		HttpURLConnection conn = null;
		boolean doNotClose = false;
		try {
			switch(method) {
				case GET: conn = createGetConnection(url, query); break;
				case PUT: conn = createPutConnection(url, query); break;
				case DELETE: conn = createDeleteConnection(url, query); break;
				case POST: conn = createPostConnection(url, query); break;
				default: throw new APIConnectionException(String.format("Unknown HTTP method %s."));
			}
			
			int responseCode = conn.getResponseCode(); //triggers the request
			if (responseCode >= 200 && responseCode < 300) 
			{
				if ("application/json".equals(conn.getContentType()))
				{
					String body = getResponseBody(conn.getInputStream());
					return new APIResponse(responseCode, body);
				}
				else
				{
					doNotClose = true;
					return new APIResponse(responseCode, conn.getInputStream());
				}
			}
			else
				return new APIResponse(responseCode, getResponseBody(conn.getErrorStream()));
			
		} catch (IOException e) {
			throw new APIConnectionException(String.format("Could not connect to server (%s). Please check your internet connectivity.", url));
		} finally {
			if (!doNotClose)
				if (conn != null) { conn.disconnect(); }
		}
	}

	protected static void handleAPIError(String responseBody, int responseCode) throws ServerSideException {
		
		//System.out.println(responseCode + " -> Error Response: " + responseBody);
		Status status = APIRequest.gson.fromJson(responseBody, Status.class);
		switch(responseCode) {
			case 400: throw new APIErrorException(status); 
			case 401: throw new AuthenticationException(status.getApiResult().getStatus());
			case 404: throw new APIException(status.getApiResult().getStatus());
			case 503: throw new APIException(status.getApiResult().getStatus());
			default: throw new APIException(status.getApiResult().getStatus());
		}
	}
	
	private static class CustomizedHostnameVerifier implements HostnameVerifier {
		public boolean verify(String hostname, SSLSession session) {
			return true;
		}
	}
	
	/**
     * Trust every server - don't check for any certificate. This should only be called if server certificate is invalid
     */
    public static void trustAllHosts() {
          // Create a trust manager that does not validate certificate chains
          TrustManager[] trustAllCerts = new TrustManager[] { new X509TrustManager() {
                  public java.security.cert.X509Certificate[] getAcceptedIssuers() {
                          return new java.security.cert.X509Certificate[] {};
                  }

                  public void checkClientTrusted(X509Certificate[] chain,
                                  String authType) throws CertificateException {
                  }

                  public void checkServerTrusted(X509Certificate[] chain,
                                  String authType) throws CertificateException {
                  }
          } };

          // Install the all-trusting trust manager
          try {
                  SSLContext sc = SSLContext.getInstance("TLS");
                  sc.init(null, trustAllCerts, new java.security.SecureRandom());
                  HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
          } catch (Exception e) {
                  e.printStackTrace();
          }
      }	
}
